// ABCDswitch - This is a DIVERSE DSO which loads sets of files specified 
// by an environment variable, DPF_ABCDswitch_FILES. Each set has the same
// number of files, N, which the program determines. A set of buttons
// in shared memory determines which sets of files are displayed. Bits
// are turned on with the buttons for each set of files.
//
// Authors:
//     Adele Peskin
//      Scientific Applications & Visualization Group
//           SAVG web page: http://math.nist.gov/mcsd/savg/
//      Mathematical & Computational Science Division
//           MCSD web page: http://math.nist.gov/mcsd/
//      National Institute of Standards and Technology
//           NIST web page: http://www.nist.gov/
//
//    Steve Satterfield (SAVG) and John Kelso (Virginia Tech) provided 
//    the idea for the program
//       Virginia Tech, http://www.diverse.vt.edu/, http://www.cave.vt.edu/
//
//
// For information about this DSO, 
// contact Adele Peskin, peskin@boulder.nist.gov
//
//
// This code is a modification to the example generic dtkAugment DSO provided
// with the DIVERSE distribution. The following disclaimer applies to the NIST
// modifications.
// 
// "This software was developed at the National Institute of Standards
// and Technology by employees of the Federal Government in the course
// of their official duties. Pursuant to title 17 Section 105 of the
// United States Code this software is not subject to copyright
// protection and is in the public domain.  xxxxxxx is an experimental 
// system.  NIST assumes no responsibility whatsoever for its use by 
// other parties, and makes no guarantees, expressed or implied, about 
// its quality, reliability, or any other characteristic.
// 
// We would appreciate acknowledgement if the software is used."

#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

class ABCDswitch : public dtkAugment {
public:
  ABCDswitch(dpf *app) ;
  ~ABCDswitch(void) ;
  int postConfig(void);
  int postFrame(void);
  
private:
  dpf* app;
  dpfDisplay *display;
  dpfToggle** togNodes;
  int cnt;
  int filecnt;
  unsigned char* toggleChars;
  char** file_Lists;
  char** file_names;
  pfNode** objectPtrs;
  pfGroup* rootGroup;
  bool testNames(char* name1, char* name2);
  void findFilesPerLine(char* line);
  int checkFilesPerLine(char* line);
  void breakLineIntoFiles(char* line);
  void processFilename(char* nameStore, int wordcnt);
  pfNode* readIfNew(char* inName);
  int testSuffix(char* nameStore);
  dtkSharedMem* shm;
  signed char ButtonVal;
  int filesPerLine;
  int numberOfLines;
  int charSize;
} ;

// create object and set description.
// be sure to validate it!
ABCDswitch::ABCDswitch(dpf *app_in) :
  dtkAugment("ABCDswitch")
{
  setDescription("loads files specified by DPF_ABCDSWITCH_FILES"
		 " environment variable and display one of N"
		 " sets of data files") ;
  app = app_in;
  validate();
  cnt = 0 ;
  filecnt = 0;
  file_Lists = NULL;
  toggleChars = new unsigned char[100];
  charSize = 100;
  file_names = NULL;
  objectPtrs = NULL;
  filesPerLine = 0;
  numberOfLines = 0;
  ButtonVal = 0;
  //  Check if there is a segment "BUTTONVAL" to control a transformation
  shm = new dtkSharedMem("BUTTONVAL");
  if(!shm || shm->isInvalid())
      shm = NULL ;
} ;

// destructor
ABCDswitch::~ABCDswitch(void) {
}


int ABCDswitch::postConfig(void) {
 
  // first get the display and make and set the group node
  display = app->display();
  rootGroup = new pfGroup;
  display->world()->addChild(rootGroup);
  
  // init file loaders
  char *env = getenv("DPF_ABCDSWITCH_FILES") ;
  char *original_list ;
  char line[2047];

  if (env) {
    original_list = strdup(env) ;
    // strtok() is broken in GNU libc, so we roll our own here.
    char *end = original_list ;
    for(;*end;) {
      for(; *end != ':' && *end; end++);
      if(*end) { *end = '\0'; end++; }
      // add to the list of files to read
      if (strlen(original_list)>0) {
	// read each file and fill the filea and fileb arrays with names
	FILE *fp = fopen(original_list, "r");        
        if (!fp) {
	  dtkMsg.add(DTKMSG_ERROR,
		     "ABCDswitch::preConfig: error opening input file %s\n",
		     original_list) ;
	  return DTK_CALLBACK_ERROR;
        }
	
	printf("next file off of the list \"%s\"\n",original_list) ;
        // in each file, lines containing sets of files
        while ((fscanf(fp,"%s",line)==1)) {
	  // if this is the first line: find the number of files per line
          if (filesPerLine < 1) {
            // copy the line and find the number of files per line
            char* firstline = strdup(line);
	    findFilesPerLine(firstline);
	  }
	  else {
	    // check the line to make sure it has the right number of files
            char* nextline = strdup(line);
	    int num = checkFilesPerLine(nextline);
            if (num != filesPerLine) {
	      dtkMsg.add(DTKMSG_ERROR,
		     "ABCDswitch::preConfig: bad line: not enough files%s\n",
		     line) ;
	      continue;
	    }
	  }

	  // break up the line into filesPerLine file names and assign a toggle 
	  // and set a toggle for each
          breakLineIntoFiles(line);
	  original_list = end;
	}
      }
    }
  }

  // read setting for whether to load transformation from outside file
  if (shm) {
    shm->read(&ButtonVal);
  }

  return DTK_REMOVE_CALLBACK;
}


bool ABCDswitch::testNames(char* name1, char* name2) {
   bool sameName = true;
   int next = 0;
   for ( int j = 0; j < 10; ++j ) {
     if (name1[next] != name2[next]) {
       sameName = false;
       return sameName;
     }
     ++next;
   }
   return sameName;
}

void ABCDswitch::findFilesPerLine(char* line) {
   int len = strlen(line);
   for ( int i = 0; i < len; i++ ) {
     if (line[i] == ':') ++filesPerLine; 
   }
   // add an extra file after the last colon
   if (filesPerLine > 0) ++filesPerLine;
}

int ABCDswitch::checkFilesPerLine(char* line) {
   int len = strlen(line);
   int num = 0;
   for ( int i = 0; i < len; i++ ) {
     if (line[i] == ':') ++num; 
   }
   // add an extra file after the last colon
   if (filesPerLine > 0) ++num;
   return num;
}

void ABCDswitch::breakLineIntoFiles(char* line) {
   int wordcnt = filesPerLine-1;
   char* line2 = strdup(line);

   // set up a togglechar for this line
   toggleChars[numberOfLines] = 1;

   char *colon;
   char* file_a;
   char* file_b;

   for ( int k = 0; k < filesPerLine-1; k++ ) {
     colon = strrchr(line2, ':') ;
     *colon = '\0' ;
     file_a = strdup(line2) ;
     file_b = strdup(colon+1) ;

     // file are stripped off the end
     // tog number must be in reverse order
     processFilename(file_b, wordcnt);
     line2 = file_a;
     --wordcnt;
   }
   // process the last one
   processFilename(file_a, wordcnt);

   // increment the number of files in each set
   ++numberOfLines;
}

void ABCDswitch::processFilename(char* nameStore, int togcnt) {
  dpfToggle* tog = new dpfToggle(&toggleChars[numberOfLines], togcnt);

  //read in new object (inName) and load if new
  pfNode* newObject = readIfNew(nameStore);

  // add the object to the toggle
  tog->addChild(newObject);

  // add the tog to the root node and the toggle list
  rootGroup->addChild(tog);
  ++cnt;
  if (cnt > charSize) {
    charSize += 100;
    toggleChars = (unsigned char *) realloc(toggleChars, sizeof(unsigned char)*(charSize));
  }
}

pfNode* ABCDswitch::readIfNew(char* inName) {
  pfNode* newObject;
  // see if this filename is new
  bool isItNew = true;
  int place = -1;
  for ( int i = 0; i < filecnt; i++ ) {
    if (testNames(inName,file_names[i])) {
      isItNew = false;
      place = i;
    }
  }

  if (isItNew) {
    // load the file into a pfNode
    newObject = pfdLoadFile(inName);
    file_names = (char **) realloc(file_names, sizeof(char*)*(filecnt+1));
    // make space for the new file name in file_names
    file_names[filecnt] = inName;
    objectPtrs = (pfNode **) realloc(objectPtrs, sizeof(pfNode*)*(filecnt+1));
    objectPtrs[filecnt] = newObject;
    ++filecnt;
  }
  else {
    //pull out the old node pointer
    newObject = objectPtrs[place];
  }

  // load the converter if necessary
  char *tempa = inName;
  int err = testSuffix(tempa);
  return newObject;
} 

int ABCDswitch::testSuffix(char* nameStore) {
  int sufc = 0;
  char *suf ;
  char **sufv = NULL ;
  if (strlen(nameStore)>0) {
    suf = strrchr(nameStore, '.') ;
    if (suf) {
      int suf_found = 0 ;
      for (int c = 0; c<sufc; c++) {
	if (!strcmp(suf,sufv[c])) {
	  suf_found = 1 ;
	  break ;
		    
	}
      }
      if (!suf_found) {
	sufv = (char **) realloc(sufv, sizeof(char *)*(sufc+1));
	sufv[sufc] = strdup(suf) ;
	sufc++ ;
	if(!pfdInitConverter(suf)) {
	  dtkMsg.add(DTKMSG_ERROR,
		     "ABCDswitch::preConfig: pfdInitConverter(%s) failed.\n",
		     suf) ;
	  return DTK_CALLBACK_ERROR;
	}
      }
    } else {
      dtkMsg.add(DTKMSG_ERROR,
		 "ABCDswitch::preConfig: no file suffix found for file \"%s\"\n",nameStore) ;
      return DTK_CALLBACK_ERROR;
    }
	      
    if(!pfdInitConverter(suf)) {
      dtkMsg.add(DTKMSG_ERROR,	 "pfdInitConverter(%s) failed.\n",
		 suf) ;
      return DTK_CALLBACK_ERROR;
    }      
  }
  return 0;
}


// called every display loop, as a pfFrame callback, after pfDraw()
// this return value causes it to only be called once, and the object removed
int ABCDswitch::postFrame(void) {

  // reread button for new transformation values
  if (shm) {
     shm->read(&ButtonVal);
     
     // set each toggle from ButtonVal
     for ( int j = 0; j < numberOfLines; j++ ) {
       toggleChars[j] = ButtonVal;
     }
  }

  return DTK_CONTINUE ;
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new ABCDswitch(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
